#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

install_requires = \
['signature_dispatch']

extras_require = \
{'doc': ['sphinx', 'sphinx_rtd_theme', 'autoclasstoc'],
 'test': ['pytest', 'pytest-cov', 'coveralls']}

setup(name='autoprop',
      version='3.0.0',
      description='Infer properties from accessor methods.',
      author='Kale Kundert',
      author_email='kale@thekunderts.net',
      url='https://github.com/kalekundert/autoprop',
      py_modules=['autoprop'],
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='~=3.6',
     )
